# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2015,2017
"""
SPL primitive operators that call a Python function or
callable class are created by decorators provided :py:mod:`streamsx.spl.spl`
Once created the operators become part of a toolkit and may be used
like any other SPL operator.
"""
